=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed June 2008 by Fredo6

# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   BzRoundEdge.rb
# Original Date	:   10 June 2008 - version 1.0
# Type			:   Sketchup Tools
# Description	:   Round Edges in 3D, including at corners
# Menu Items	:   Tools --> "Round Edges"
# Toolbar		:   Name = "RoundEdge"
# Context Menu	:   yes
# Usage			:   See Tutorial and Quick Ref Card in PDF format
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

require 'sketchup.rb'
require 'LibTraductor.rb'			# for language translation

module BzRoundEdge

RDE_DIR = "RDE_Dir"
RDE_TOOLBAR = "RoundEdge"

#Constants for LineOnSurface Module (do not translate)	
ICON_BEZIER = "RoundEdge_Bezier"
CURSOR_BEZIER_OK = "RoundEdge_Bezier_OK"
CURSOR_BEZIER_NO = "RoundEdge_Bezier_NO"

RDE_TITLE_SUBMENU = ["Round Edges",
                     "|FR| Arrondisseur de segments"]
					 
MSG_RoundEdge_Bezier = ["Round Edges by Bezier",
	                   "|FR| Arrondisseur d'ar\tes (Bezier)"]		
					   
MSG_RoundEdge_Origin = ["Pick Edges whith Mouse depressed",
                        "|FR| S\lectionner des ar\tes avec la souris appuy\e"]
MSG_Label_Edge = ["Edges", "|FR| Ar\tes"]	

DLG_EnumYesNo = { 'Y' => "Yes |FR| Oui", 
				  'N' => "No |FR| Non" }			 

DLG_Distance = ["Distance", "|FR| Distance"]
DLG_NbSeg = ["Number of segments", "|FR| Nombre de segments"]
DLG_GroupOffset = ["Offset Red Axis for Group (0 means no Group)", 
                   "|FR| Decalage Axe Rouge pour le Groupe (0 pour pas de groupe)"]
DLG_Smooth = ["Smoothen round shapes", 
              "|FR| Lissage des arrondis"]
DLG_Soft = ["Soften junction with round shapes", 
            "|FR| Adoucir bordures"]
			   

RDE_Edge = Struct.new("RDE_Edge", :edge, :ee1, :ee2, :face1, :face2, :vecedge,
                                  :ptmid1, :ptmid2, :bzpts)
RDE_EdgeEnd = Struct.new("RDE_EdgeEnd", :edge, :vertex, :vd, :ed, :edges_add, :edges_erase,
                                        :vecedge, :ptbord1, :ptbord2, :edge1, :edge2, :biplane1, :biplane2,
										:stop_plane, :profile, :lisec, :touched, :draw_end, :part)
RDE_Vertex = Struct.new("RDE_Vertex", :vertex, :ledges, :list_ee)
				 
#--------------------------------------------------------------------------------------------------------------
# Top Calling functions: create the classes and launch the tools
#--------------------------------------------------------------------------------------------------------------			 				   

def BzRoundEdge.launch
	@tool_roundedge = ToolRoundEdge.new unless @tool_roundedge
	Sketchup.active_model.select_tool @tool_roundedge
end

def BzRoundEdge.check_selection
	ss = Sketchup.active_model.selection
	ledge = []
	ss.each do |e|
		ledge.push e if e.class == Sketchup::Edge && BzRoundEdge.edge_valid?(e)
	end	
	
	ledge
end
	
def BzRoundEdge.edge_valid?(edge)
	return false unless edge.faces.length == 2
	return false if edge.soft? || edge.smooth? || edge.hidden?
	face1 = edge.faces[0]
	face2 = edge.faces[1]
	return (face1.normal.parallel?(face2.normal)) ? false : true
end
	
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ToolRoundEdge: Tool to erase lines (plain or construction) on a surface
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
class ToolRoundEdge
					
def initialize
	#Loading strings and cursors
	Traductor.load_translation BzRoundEdge, /MSG_/, binding, "@msg_"
	@idcursor_OK = create_cursor CURSOR_BEZIER_OK, 3, 3
	@idcursor_NO = create_cursor CURSOR_BEZIER_NO, 3, 3
	@tr_id = Geom::Transformation.new

	#initializing variables
	@ip = Sketchup::InputPoint.new	
end

def activate
	@model = Sketchup.active_model
	@selection = @model.selection
	@entities = @model.active_entities
	@tr_id = Geom::Transformation.new

	reset_selection
	
	@list_edges = BzRoundEdge.check_selection
	@noselection = (@list_edges.length == 0)
	
	@selection.clear
	unless @noselection	
		@selection.add @list_edges
		execute_drawing
		reset_selection
	end	
	
	@enter_down = false
	reset_selection
	@selection.clear
	@button_down = false
	
	info_show
end

def reset_selection
	@list_edges = []
	@list_valid = []
	@list_pts = []
	@hsh_vertices = {}
	@hsh_edges = {}
end

def deactivate(view)
	view.invalidate
end

#----------------------------------------------------------------------
#Algorithm section for the RoundEdge script
#----------------------------------------------------------------------

#Constructing the list and hash table of vertices
def construct_all_vertices
	#constructing the list of vertices
	@list_edges.each do |edge|
		construct_vertex edge.start
		construct_vertex edge.end
	end	
		
	#Constructing the list of edges
	@hsh_vertices.each do |key, vd|
		vd.ledges.each do |e|
			construct_edge(vd, e)
		end
	end	

end

#Constructing a single vertex
def construct_vertex(vertex)
	return if @hsh_vertices[vertex.to_s]
	vd = RDE_Vertex.new
	@hsh_vertices[vertex.to_s] = vd	
	vd.vertex = vertex
	vd.ledges = []
	vd.list_ee = []
	vertex.edges.each do |e|
		next unless @list_edges.include?(e)
		vd.ledges.push e
	end	
end

def compute_bezier_profiles(ed)
	ee1 = ed.ee1
	ee2 = ed.ee2
	edge = ed.edge
	ptmid = Geom.linear_combination 0.5, edge.start.position, 0.5, edge.end.position
	
	face1 = edge.faces[0]
	vec = RDEG.normal_in_to_edge edge, face1
	ed.ptmid1 = ptmid.offset vec, @distance
	
	face2 = edge.faces[1]
	vec = RDEG.normal_in_to_edge edge, face2
	ed.ptmid2 = ptmid.offset vec, @distance
	
	if @nb_seg == 1
		ed.bzpts = [ed.ptmid1, ed.ptmid2]
	else
		ed.bzpts = BezierCurve.compute [ed.ptmid1, ptmid, ed.ptmid2], @nb_seg	
	end	
end

#Compute the delimiters on each face (to remove the corner edges)
def compute_borders_at_edge_end(ed, ee)
	vertex = ee.vertex
	edge = ed.edge
	face1 = edge.faces[0]
	face2 = edge.faces[1]
	edge1 = nil
	edge2 = nil
	
	#Finding the other cofacial edges
	vertex.edges.each do |e|
		next if e == edge
		edge1 = e if e.faces.include?(face1)
		edge2 = e if e.faces.include?(face2)
	end
	
	line = [ed.ptmid1, ed.vecedge]
	if RDEG.edge_colinear?(edge, edge1)
		ee.ptbord1 = Geom.intersect_line_plane line, [vertex.position, ee.vecedge]
		ee.edges_add.push [vertex.position, ee.ptbord1]
		ee.draw_end = true
	elsif @list_edges.include?(edge1)
		ee.biplane1 = RDEG.compute_bissector_plane vertex, edge, edge1
		ee.ptbord1 = Geom.intersect_line_plane line, ee.biplane1
	else
		ee.ptbord1 = Geom.intersect_line_line line, edge1.line
		ee.edges_erase.push [vertex.position, ee.ptbord1]
	end
	ee.edge1 = (@list_edges.include?(edge1)) ? edge1 : nil
	
	line = [ed.ptmid2, ed.vecedge]
	if RDEG.edge_colinear?(edge, edge2)
		ee.ptbord2 = Geom.intersect_line_plane line, [vertex.position, ee.vecedge]
		ee.edges_add.push [vertex.position, ee.ptbord2]
	elsif @list_edges.include?(edge2)
		ee.biplane2 = RDEG.compute_bissector_plane vertex, edge, edge2
		ee.ptbord2 = Geom.intersect_line_plane line, ee.biplane2
	else
		ee.ptbord2 = Geom.intersect_line_line line, edge2.line
		ee.edges_erase.push [vertex.position, ee.ptbord2]
	end
	ee.edge2 = (@list_edges.include?(edge2)) ? edge2 : nil
	
	#computing the stop plane
	if ee.edge1 == ee.edge2
		ee.stop_plane = [vertex.position, ee.vecedge]
	else
		vec1 = edge1.start.position.vector_to edge1.end.position
		vec2 = edge2.start.position.vector_to edge2.end.position		
		ee.stop_plane = [vertex.position, vec1 * vec2]
	end	
end

#calculate the profile at extremity for simple connection
def compute_stop_profile(ee)
	vec = ee.vecedge
	ee.lisec = []
	bzpts = ee.ed.bzpts
	nb = bzpts.length - 2
	for i in 0..nb
		p1 = bzpts[i]
		p2 = Geom.intersect_line_plane([p1, vec], ee.stop_plane)
		p4 = bzpts[i+1]
		p3 = Geom.intersect_line_plane([p4, vec], ee.stop_plane)
		ee.lisec[i] = [p1, p2, p3, p4]
	end	
end

#constructing a single edge
def construct_edge(vd, edge)
	#creating the Edge structure if not created yet
	vertex = vd.vertex
	origin = vertex.position
	ed = @hsh_edges[edge.to_s]
	unless ed
		ed = RDE_Edge.new
		@hsh_edges[edge.to_s] = ed
		ed.edge = edge
		ed.vecedge = origin.vector_to edge.other_vertex(vertex).position
		compute_bezier_profiles ed
		ee = ed.ee1 = RDE_EdgeEnd.new
	else
		ee = ed.ee2 = RDE_EdgeEnd.new
	end	
	ee.edge = edge
	ee.ed = ed
	ee.vertex = vertex
	ee.vd = vd
	vd.list_ee.push ee
	ee.vecedge = edge.other_vertex(vertex).position.vector_to origin
	ee.edges_add = []
	ee.edges_erase = []
	ee.touched = false
	ee.draw_end = false
	ee.part = []
	
	#Computing the bording lines
	compute_borders_at_edge_end ed, ee
	unless ee.edge1 || ee.edge2
		compute_stop_profile ee
	else	
		compute_initial_lisec ee
	end	
end

#Calculate a junction of rounding at an extremity - Only used for 3 or more edges termination at the vertex
def compute_intersect_at_vertex(entities, ed, ee)
	return unless ee.edge1 || ee.edge2
	
	#gathering the list of edges ends
	vd = ee.vd
	ed = ee.ed
	list_ee = []
	vd.list_ee.each { |eee| list_ee.push eee if eee != ee }
	
	#Only 2 edges - Bypass treatment
	if list_ee.length == 1
		#compute_twin_intersect ee
		#return
	end
	
	#loop of other ends
	list_ee.each do |eeother|
		#create the working group
		g = entities.add_group
		
		#generate the curved contour for the other edge_end
		compute_bezier_half_faces(g.entities, eeother)
		
		#Loop of each individual face and calculation of intersection
		nb = ed.bzpts.length - 2
		for i in 0..nb
			intersect_blade(g.entities, ee, i)
		end	
		
		#Clear the group
		entities.erase_entities g
	end
end

#Simple junction with only 2 edges
def compute_twin_intersect(ee)
	ed = ee.ed
	if ee.edge1
		plane = RDEG.compute_bissector_plane ee.vertex, ed.edge, ee.edge1
	else
		plane = RDEG.compute_bissector_plane ee.vertex, ed.edge, ee.edge2
	end
	bzpts = ed.bzpts
	vec = ee.vecedge
	
	nb = bzpts.length - 2
	ee.lisec = []
	for i in 0..nb
		p1 = bzpts[i]
		p2 = Geom.intersect_line_plane [bzpts[i], vec], plane
		p3 = Geom.intersect_line_plane [bzpts[i+1], vec], plane
		p4 = bzpts[i+1]
		ee.lisec[i] = [p1, p2, p3, p4]
	end	
end

#Compute the initial profiles at intersection of edges
def compute_initial_lisec(ee)
	ed = ee.ed
	edge = ed.edge
	leng = edge.length
	bzpts = ed.bzpts
	vec = ee.vecedge
	
	nb = bzpts.length - 2
	ee.lisec = []
	for i in 0..nb
		p1 = bzpts[i]
		p2 = bzpts[i].offset vec, leng
		p3 = bzpts[i+1].offset vec, leng
		p4 = bzpts[i+1]
		ee.lisec[i] = [p1, p2, p3, p4]
	end	
end

#generate the pipe faces at edge end
def compute_bezier_half_faces(entities, ee)
	ed = ee.ed
	edge = ed.edge
	leng = edge.length
	bzpts = ed.bzpts
	vec = ee.vecedge
	
	nb = bzpts.length - 2
	for i in 0..nb
		p1 = bzpts[i]
		p2 = bzpts[i].offset vec, leng
		p3 = bzpts[i+1].offset vec, leng
		p4 = bzpts[i+1]
		entities.add_face [p1, p2, p3, p4]
	end	
end

#compute the intersection of a blade with anothr half-pipe
def intersect_blade(entities, ee, i)
	ed = ee.ed
	leng = ed.edge.length
	bzpts = ed.bzpts
	vec = ee.vecedge
	ptref = bzpts[i]

	#generating the blade
	gface = entities.add_group
	face_blade = gface.entities.add_face ee.lisec[i]
	
	#intersecting the blade within the group
	entities.intersect_with false, @tr_id, gface.entities, @tr_id, true, face_blade
		
	#Finding the right face to keep
	goodface = nil
	gface.entities.each do |face|
		next unless face.class == Sketchup::Face
		if RDEG.point_is_face_vertex(ptref, face)
			goodface = face
			break
		end	
	end
	
	#storing new profile
	if goodface
		ee.lisec[i] = []
		goodface.outer_loop.vertices.each do |v|
			ee.lisec[i].push v.position
		end
	end

	entities.erase_entities gface	
end


#Re-arrange a suite of points, from ptref1 to ptref2, removing them from the contour
def reorder_profile(pts, ptref1, ptref2)
	nb = pts.length - 1
	return pts[0..-3].reverse if pts.last == ptref2 && pts[-2] == ptref1
	return pts[0..-3] if pts.last == ptref1 && pts[-2] == ptref2
	return pts[2..-1].reverse if pts[0] == ptref1 && pts[1] == ptref2
	return pts[2..-1] if pts[0] == ptref2 && pts[1] == ptref1
	return pts[1..-2] if pts[0] == ptref1
	return pts[1..-2].reverse if pts[0] == ptref2

	order = 0
	partA = []
	partB = []
	part = partA
	for i in 0..nb
		if (pts[i] == ptref1)
			order = 1 if (order == 0) 
			part = partB
		elsif (pts[i] == ptref2)
			order = -1 if (order == 0)
			part = partB
		else
			part.push pts[i]
		end
	end	
	return (order == -1) ? partB + partA : partA.reverse + partB.reverse
end

#Generate the final blade faces
def compute_final_faces(entities, ed)
	profile1 = ed.ee1.lisec
	nb1 = profile1.length - 1
	profile2 = ed.ee2.lisec
	nb2 = profile2.length - 1
	return if nb1 < 0 || nb2 < 0
	bzpts = ed.bzpts
	lfaces = []

	#Reordering profiles and generating the faces for each blade
	for i in 0..nb1
		ptref1 = bzpts[i]
		ptref2 = bzpts[i+1]
		part1 = reorder_profile(profile1[i], ptref1, ptref2)
		part2 = reorder_profile(profile2[i], ptref1, ptref2)
		part2 = part2.reverse
		pts = part1 + part2	
		ed.ee1.part.push part1		
		ed.ee2.part.push part2
		lfaces.push entities.add_face(pts) if pts.length > 2
	end	
	
	#Transfering properties
	face1 = ed.edge.faces[0]
	face2 = ed.edge.faces[1]
	nb = lfaces.length - 1
	nblim = nb / 2
	norm1 = face1.normal
	norm2 = face2.normal

	for i in 0..nblim
		newface = lfaces[i]
		newface.reverse! if newface.normal % norm1 < 0
		norm1 = newface.normal
		RDEG.transfer_face face1, newface
		
		newface = lfaces[nb-i]
		newface.reverse! if newface.normal % norm2 < 0
		norm2 = newface.normal
		RDEG.transfer_face face2, newface
	end
	
	lfaces.each do |face|
		face.edges.each do |e|
			e.soft = true
			e.smooth = @option_smooth
		end
	end	
end

def draw_end(entities, ee)
	part = ee.part
	pts = [part[0].first]
	part.each do |lpt|
		pts += lpt [1..-1]
	end
	le = entities.add_edges pts
	le.each do |e| 
		next unless e.class == Sketchup::Edge
		e.smooth = @option_smooth
		e.soft = @option_soft
	end	
end

#Create the Geometry
def execute_drawing
	
	#Asking for parameters with dialog box
	return unless check_param
	
	#Constructing all structures for calculation
	construct_all_vertices
	
	pbar = ProgressionBar.new @hsh_edges.length, @msg_Label_Edge
	
	@model.start_operation Traductor[@msg_RoundEdge_Bezier]
	
	grp = @entities.add_group
	entities = grp.entities
	@hsh_edges.each do |key, ed|

		#computing borders on each face (to erase the edges)
		g = @entities.add_group	
		e = g.entities.add_edges ed.ee1.ptbord1, ed.ee2.ptbord1
		e[0].smooth = @option_smooth
		e[0].soft = @option_soft
		e = g.entities.add_edges ed.ee1.ptbord2, ed.ee2.ptbord2
		e[0].smooth = @option_smooth
		e[0].soft = @option_soft
		ed.ee1.edges_add.each { |seg| g.entities.add_edges seg }
		ed.ee2.edges_add.each { |seg| g.entities.add_edges seg }
		g.explode
		
		#Intersecting faces
		compute_intersect_at_vertex(entities, ed, ed.ee1) #if ed.ee1.edge1
		compute_intersect_at_vertex(entities, ed, ed.ee2) #if ed.ee2.edge1
		
		#Final faces
		compute_final_faces entities, ed
		
		pbar.countage
				
	end
	
	#erasing the original corner edges
	unless @option_group
		@hsh_edges.each do |key, ed|
			@entities.erase_entities ed.edge
			#draw_end @entities, ed.ee1 if ed.ee1.draw_end
			#draw_end @entities, ed.ee2 if ed.ee2.draw_end
		end
		grp.explode
	else
		vec = Geom::Vector3d.new @group_offset, 0, 0
		tt = Geom::Transformation.translation vec
		@entities.transform_entities tt, grp
	end	
	
	@model.commit_operation
end

#Invoke dialog box to get current parameters for the Operation
def check_param
	unless @dlg
		dsnap = RDEG.dsnap.to_l
		distdef = 50.cm
		distmin = 0.5.cm
		@hsh_params = {}
		@dlg = Traductor::DialogBox.new @msg_RoundEdge_Bezier 
		@dlg.field_numeric "Distance", DLG_Distance, distdef.to_l, distmin.to_l, nil 
		@dlg.field_numeric "NbSeg", DLG_NbSeg, 8, 1, 30 
		@dlg.field_numeric "GroupOffset", DLG_GroupOffset, 0.cm, nil, nil 
		@dlg.field_enum "Smooth", DLG_Smooth, 'Y', DLG_EnumYesNo
		@dlg.field_enum "Soft", DLG_Soft, 'Y', DLG_EnumYesNo
	end	

	#Invoking the dialog box
	return false unless @dlg.show! @hsh_params		

	@distance = @hsh_params["Distance"]
	@nb_seg = @hsh_params["NbSeg"]
	@group_offset = @hsh_params["GroupOffset"]
	@option_group = (@group_offset != 0)
	@option_smooth = (@hsh_params["Smooth"] == 'Y') ? true : false
	@option_soft = (@hsh_params["Soft"] == 'Y') ? true : false
	true
end

#----------------------------------------------------------------------
#Graphical section for the Tool methods
#----------------------------------------------------------------------

def onCancel(flag, view)
	#User did an Undo
	case flag
	when 1, 2	#Undo or reselect the tool
		activate
		return
	when 0	#user pressed Escape
		reset_selection
		@selection.clear
	end
end

def create_cursor(cursorname, hotx=0, hoty=0)
	cursorfile = "RDE_cursor_" + cursorname + ".png"
	cursorpath = Sketchup.find_support_file cursorfile, "Plugins/" + RDE_DIR
	cursorpath = Sketchup.find_support_file cursorfile, "Plugins" unless cursorpath
	(cursorpath) ? UI::create_cursor(cursorpath, hotx, hoty) : 0
end

def onSetCursor
	UI::set_cursor (@ok) ? @idcursor_OK : @idcursor_NO
end

def getMenu(menu)
	if (@list_edges.length > 0)
		menu.add_item(@msg_MnuDone) { execute_drawing }
		menu.add_separator
	end	
	option_contextual_menu menu
	true
end

#Populate the options in the Contextual menu
def option_contextual_menu(menu)	

end

def onLButtonDown(flags, x, y, view)
	@button_down = true
	onMouseMove(flags, x, y, view)
end

def onLButtonUp(flags, x, y, view)
	execute_drawing
	reset_selection
	@button_down = false
end

#Key Up
def onKeyUp(key, rpt, flags, view)
	key = Traductor.check_key key, flags, true

	case key
		#Toggling between fixed and variable length
		when COPY_MODIFIER_KEY
			view.invalidate
			info_show
	end	
	@control_down = false
end

#Key down
def onKeyDown(key, rpt, flags, view)
	key = Traductor.check_key key, flags, false

	case key			
		#Calling options
		when CONSTRAIN_MODIFIER_KEY
			onMouseMove 2, @x, @y, view if @button_down

		when COPY_MODIFIER_KEY
			@control_down = true
			return
			
		when 13
			@enter_down = true
			@control_down = false
			return
		else
			@control_down = false
			return
			
	end	
	@control_down = false
	
	view.invalidate
	info_show
end

#Check if the edges are already part of the selection
def already_selected(ledge)
	status = false
	ledge.each do |e|
		if @list_edges.include?(e) 
			status = true
		elsif BzRoundEdge.edge_valid?(e)
			@list_edges.push e
			@selection.add e
			status = true
		end	
	end
	status
end

#Check if edges are already known
def already_valid(ledge)
	status = false
	ledge.each do |e|
		if @list_valid.include?(e) 
			status = true
		elsif BzRoundEdge.edge_valid?(e)
			@list_valid.push e
			status = true
		end	
	end
	status
end

#Check and mark edge or edges at vertex for selection
def mark_edge_for_selection(edge, vertex)
	le = (vertex) ? vertex.edges : [edge]
	(@button_down) ? already_selected(le) : already_valid(le)
end

#Mouse Move method
def onMouseMove(flags, x, y, view)
	@x = x
	@y = y
	@ok = false
	ph = view.pick_helper
	ph.do_pick x, y
	edge = ph.best_picked	
	return unless edge && edge.class == Sketchup::Edge
	@ip.pick view, x, y
	@ok = mark_edge_for_selection edge, @ip.vertex
	if (@ok) && edge 
		@pk_edge = edge
	else
		@pk_edge = nil
	end	
	onSetCursor
	view.invalidate
end	

#Draw method for Polyline tool
def draw(view)	
	if @button_down == false && @pk_edge && @pk_edge.valid?
		view.drawing_color = 'orange'
		view.line_stipple = ""
		view.line_width = 4
		view.draw GL_LINE_STRIP, @pk_edge.start.position, @pk_edge.end.position
	end
end

#display information in the Sketchup status bar
def info_show
	msg = @msg_RoundEdge_Bezier 
	msg += " - " + @msg_RoundEdge_Origin
	
	Sketchup.set_status_text msg	
end

end	#End Class ToolRoundEdge

#--------------------------------------------------------------------------------------------------------------
# Bezier methods
#--------------------------------------------------------------------------------------------------------------			 				   

class BezierCurve

# Evaluate the curve at a number of points and return the points in an array
def BezierCurve.compute(pts, numpts) 
    curvepts = []
    dt = 1.0 / numpts

    # evaluate the points on the curve
    for i in 0..numpts
        t = i * dt
        curvepts[i] = BezierCurve.evaluate(pts, t)
    end 
    curvepts
end

def BezierCurve.evaluate(pts, t)
    degree = pts.length - 1
    if degree < 1
        return nil
    end
    
    t1 = 1.0 - t
    fact = 1.0
    n_choose_i = 1

    x = pts[0].x * t1
    y = pts[0].y * t1
    z = pts[0].z * t1
    
    for i in 1...degree
		fact = fact*t
		n_choose_i = n_choose_i*(degree-i+1)/i
        fn = fact * n_choose_i
		x = (x + fn*pts[i].x) * t1
		y = (y + fn*pts[i].y) * t1
		z = (z + fn*pts[i].z) * t1
    end

	x = x + fact*t*pts[degree].x
	y = y + fact*t*pts[degree].y
	z = z + fact*t*pts[degree].z

    Geom::Point3d.new(x, y, z)  
end 

end	#class BezierCurve

#--------------------------------------------------------------------------------------------------------------
# Class ProgressionBar: progress bar in the Sketchup Status text area
#--------------------------------------------------------------------------------------------------------------
class ProgressionBar

#Initialization of progress bar
def initialize(nbelts, label)
	@pb_nbelts = nbelts
	@pb_label = label
	@pb_progression = 0
	@pb_rangemax = 200
	@pb_range = 0
	@pb_time0 = Time.now	
end

#Increment the Progression Bar by <nb> steps
def countage(nb=1)
	@pb_progression += nb
	f = 100 * @pb_progression / @pb_nbelts
	percent = f.to_i
	if (percent != @pb_range)
		@pb_range = percent
		n = 1 + percent * @pb_rangemax / 100
		Sketchup::set_status_text "|" * n.to_i
	end	
	Sketchup.set_status_text @pb_label + " #{@pb_progression} / #{@pb_nbelts}", SB_VCB_LABEL
	Sketchup::set_status_text "#{@pb_range}%  -  #{sprintf "%4.2f", Time.now - @pb_time0} sec", SB_VCB_VALUE
end

end #class ProgressionBar

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# RDEG: Utility class with standalone functions
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
class RDEG

#generic transfer for any drawing element
def RDEG.transfer_drawing_element (old_entity, new_entity)
	new_entity.layer = old_entity.layer
	new_entity.material = old_entity.material
	new_entity.visible = old_entity.visible?
	new_entity.receives_shadows = old_entity.receives_shadows?
	new_entity.casts_shadows = old_entity.casts_shadows?
end

#generic transfer for any drawing element
def RDEG.transfer_face(oldface, newface)
	RDEG.transfer_drawing_element oldface, newface
	newface.back_material = oldface.back_material
	#newface.reverse! if (newface.normal % oldface.normal < 0) 
end

#calculate the normal to an edge of a face pointing toward the outside
def RDEG.normal_in_to_edge(edge, face)
	pt1 = edge.start.position
	pt2 = edge.end.position
	vec = face.normal * pt1.vector_to(pt2)
	vec.length = 1.0
	edge.reversed_in?(face) ? vec.reverse : vec
end

#Compute plane between 2 edges - Return the normal vector
def RDEG.compute_bissector_plane(vertex, edge1, edge2)
	#Finding the common vertex
	vertex = RDEG.common_vertex edge1, edge2
	vother1 = edge1.other_vertex vertex
	vother2 = edge2.other_vertex vertex
		
	#Bissector Plane
	vec1 = vertex.position.vector_to(vother1.position).normalize
	vec2 = vertex.position.vector_to(vother2.position).normalize
	if vec1.parallel? vec2
		[vertex.position, vec1]
	else	
		[vertex.position, (vec1 * vec2) * (vec1 + vec2)]
	end	
end

#finding common vertex to 2 edges
def RDEG.common_vertex(edge1, edge2)
	edge2.vertices.each do |v|
		return v if edge1.vertices.include? v
	end	
	nil
end

#Check if 2 adjacent edges are colinear
def RDEG.edge_colinear?(edge1, edge2)
	vec1 = edge1.start.position.vector_to edge1.end.position
	vec2 = edge2.start.position.vector_to edge2.end.position
	vec1.parallel? vec2
end

def RDEG.point_is_face_vertex(pt, face)
	face.vertices.each do |v|
		return true if pt == v.position
	end
	false
end

#Retrieve the length for snapping, also used as precision
def RDEG.dsnap
	op = Sketchup.active_model.options["UnitsOptions"] 
	d = op["LengthSnapLength"]
	d.to_f
end

end	#class RDEG


#--------------------------------------------------------------------------------------------------------------
# Top Calling functions: create the class and launch the tool
#--------------------------------------------------------------------------------------------------------------			 				   

def BzRoundEdge.add_command(cmd, menutool, title, tooltip, icon_name)
	#Main command
	dir_rde = File.join "Plugins", "RDE_Dir"
	dir_main = "Plugins"
	menutool.add_item cmd
	
	cmd.status_bar_text = tooltip
	iconpath = "RDE_icon_" + icon_name + ".png"
	iconpath16 = "RDE_icon_" + icon_name + "_16.png"
	iconpath24 = "RDE_icon_" + icon_name + "_24.png"
	icon = Sketchup.find_support_file iconpath, dir_rde
	icon = Sketchup.find_support_file iconpath, dir_main unless icon
	icon16 = Sketchup.find_support_file iconpath16, dir_rde
	icon16 = Sketchup.find_support_file iconpath16, dir_main unless icon16
	icon24 = Sketchup.find_support_file iconpath24, dir_rde
	icon24 = Sketchup.find_support_file iconpath24, dir_main unless icon24
	icon16 = icon unless icon16
	icon16 = icon24 unless icon16
	icon24 = icon unless icon24
	icon24 = icon16 unless icon24
	if (icon16 || icon24)
		cmd.tooltip = tooltip
		cmd.small_icon = icon16
		cmd.large_icon = icon24
		@tlb = UI::Toolbar.new RDE_TOOLBAR unless @tlb
		@tlb.add_item cmd
	end	
end

def BzRoundEdge.manage_command
	menutool = UI.menu "Tools"
	menutool.add_separator
	submenu = menutool.add_submenu Traductor[RDE_TITLE_SUBMENU]
	
	@tlb = nil

	text = Traductor[MSG_RoundEdge_Bezier]
	cmd = UI::Command.new(text) { BzRoundEdge.launch }
	BzRoundEdge.add_command(cmd, submenu, text, text, ICON_BEZIER)
		
	@tlb.show if @tlb

	UI.add_context_menu_handler do |menu|
		if BzRoundEdge.check_selection
			menu.add_separator
			menu.add_item(Traductor[MSG_RoundEdge_Bezier]) { BzRoundEdge.launch }
		end	
	end
	
end

unless $BzRoundEdge____loaded
	BzRoundEdge.manage_command
	$BzRoundEdge____loaded = true
end
		
end	#End Module BzRoundEdge
